@@ -15,6 +15,11 @@ use Scalar::Util;
 use feature 'state';
 
 
+use constant SPACE_space        => 281;     # _vspace
+use constant SPACE_index        => 289;     # _vindex
+use constant ER_TNT_PERMISSIONS => 0x8037;
+use constant ER_TNT_SCHEMA      => 0x806D;
+
 use Mouse::Util::TypeConstraints;
 
     enum DriverType     => [ 'sync', 'async' ];
@@ -234,14 +239,14 @@ sub _invalid_schema {
 
     init:
     pause:
-    schema:
         confess "Internal error: _invalid_schema in state " . $self->state;
 
+    schema:
     connecting:
     ready:
         $self->_set_state('schema');
         $self->_reconnector->ll->send_request(select =>
-                                undef, 280, 0, [], undef, undef, 'ALL', sub {
+                                undef, SPACE_space, 0, [], undef, undef, 'ALL', sub {
             my ($state, $message, $sync) = @_;
             $self->_log(debug => 'Loading spaces');
             unless ($state eq 'OK') {
@@ -254,19 +259,29 @@ sub _invalid_schema {
             $self->_reconnector->ll->wait_response($sync, sub {
                 my ($state, $message, $resp) = @_;
                 unless ($state eq 'OK') {
-                        warn $message;
                     $self->_set_last_error([ $state, $message ]);
                     $self->_set_state('pause');
                     $cb->($state => $message);
                     return;
                 }
+                
 
-                $self->_spaces($resp->{DATA});
-                # TODO: $resp->{CODE}
+                # have no permissions
+                if ($resp->{CODE} == ER_TNT_PERMISSIONS) {
+                    $self->_spaces([]);
+                } elsif ($resp->{CODE}) {
+                    $self->_set_last_error([ ER_REQUEST =>
+                        'Can not load tarantool schema', $resp->{CODE} ]);
+                    $self->_set_state('pause');
+                    $cb->(@{ $self->last_error });
+                    return;
+                } else {
+                    $self->_spaces($resp->{DATA});
+                }
 
                 $self->_log(debug => 'Loading indexes');
                 $self->_reconnector->ll->send_request(select =>
-                    $resp->{SCHEMA_ID}, 288, 0, [], undef, undef, 'ALL', sub { 
+                    $resp->{SCHEMA_ID}, SPACE_index, 0, [], undef, undef, 'ALL', sub { 
 
                     my ($state, $message, $sync) = @_;
                     unless ($state eq 'OK') {
@@ -276,7 +291,7 @@ sub _invalid_schema {
                         return;
                     }
 
-                    # TODO: $resp->{CODE}
+
                     $self->_reconnector->ll->wait_response($sync, sub {
                         my ($state, $message, $resp) = @_;
                         unless ($state eq 'OK') {
@@ -285,7 +300,24 @@ sub _invalid_schema {
                             $cb->($state => $message);
                             return;
                         }
-                        $self->_indexes($resp->{DATA});
+                    
+                        if ($resp->{CODE} == ER_TNT_PERMISSIONS) {
+                            $self->_indexes([]);
+                        } elsif ($resp->{CODE} == ER_TNT_SCHEMA) {
+                            # collision again!
+                            $self->_invalid_schema($cb);
+                            return;
+                        
+                        } elsif ($resp->{CODE}) {
+                            $self->_set_last_error([ ER_REQUEST =>
+                                'Can not load tarantool schema', $resp->{CODE} ]);
+                            $self->_set_state('pause');
+                            $cb->(@{ $self->last_error });
+                            return;
+
+                        } else {
+                            $self->_indexes($resp->{DATA});
+                        }
 
                         $self->_set_schema($resp->{SCHEMA_ID});
                         $self->_set_state('ready');
@@ -298,7 +330,6 @@ sub _invalid_schema {
                 });
             });
         });
-
 }
 
 sub _set_schema {
@@ -501,7 +532,7 @@ sub request {
             }
 
             # schema collision
-            if ($resp->{CODE} == 0x806D) {
+            if ($resp->{CODE} == ER_TNT_SCHEMA) {
                 $self->_log(error => 'Detected schema collision');
                 $self->_log(info => 'Defer request "%s" until schema loaded', $name);
                 push @{ $self->_wait_ready } => $request;
